#include <irrlicht.h>
#include <ICursorControl.h>
#include <iostream>
#include "portalscenenode.cpp"
#include "MasterEventReceiver.cpp"
#include "crosshair.cpp"
#include "portalutils.h"

using namespace irr;
using namespace core;
using namespace scene;
using namespace video;
using namespace io;
using namespace gui;

/* Funo de movimentao da cmera. Atualiza a posio da cmera de acordo com 
os comandos do jogador. A cmera pode ser movimentada utilizando as teclas:
    W: para frente
    S: para trs
    A: para a esquerda
    D: para a direita
    Mouse: olhar
    Ao mesmo tempo, o modelo do personagem  atualizado para acompanhar os 
movimentos da cmera, tanto de translao quanto de rotao. 
    Esta funo leva em considerao o tempo decorrido entre o frame anterior e
o atual para calcular o movimento da cmera. 
    Retorna o estado de animao de personagem para que o mesmo seja atualizado.
*/

scene::EMD2_ANIMATION_TYPE movimentaCamera(scene::ICameraSceneNode* camera, scene::ISceneManager* smgr, 
                     MasterEventReceiver eventReceiver,u32 now, u32 then, ICursorControl *cursorControl, 
                     IAnimatedMeshSceneNode * nodePersonagem, scene::EMD2_ANIMATION_TYPE oldNodePersonagemAnimationState) {

	bool houveMovimento = false;
    scene::EMD2_ANIMATION_TYPE newNodePersonagemAnimationState;
    const f32 frameDeltaTime = (f32)(now - then) / 1000.f;
	const f32 MoveSpeed = 200.0;
	const f32 RotateSpeed = 100.0;
	const f32 MaxVerticalAngle = 88.0f;
    core::position2d<f32> centerCursor = core::position2d<f32>(0.5f,0.5f);
    core::position2d<f32> cursorPos = core::position2d<f32>(cursorControl->getRelativePosition().X, cursorControl->getRelativePosition().Y);
    
    core::vector3df pos = camera->getPosition();
   
    core::vector3df target = (camera->getTarget() - camera->getAbsolutePosition());
	core::vector3df relativeRotation = target.getHorizontalAngle();
	
	if (cursorControl)
	{
		if (cursorPos != centerCursor)
		{
			relativeRotation.Y -= (0.5f - cursorPos.X) * RotateSpeed;
			relativeRotation.X -= (0.5f - cursorPos.Y) * RotateSpeed;


			if (relativeRotation.X > MaxVerticalAngle*2 &&
				relativeRotation.X < 360.0f-MaxVerticalAngle)
			{
				relativeRotation.X = 360.0f-MaxVerticalAngle;
			}
			else
			if (relativeRotation.X > MaxVerticalAngle &&
				relativeRotation.X < 360.0f-MaxVerticalAngle)
			{
				relativeRotation.X = MaxVerticalAngle;
			}
		}

		cursorControl->setPosition(0.5f, 0.5f);
		centerCursor = cursorControl->getRelativePosition();
		cursorPos = centerCursor;
	}
	
	target.set(0,0, core::max_(1.f, pos.getLength()));
	core::vector3df movedir = target;

	core::matrix4 mat;
	mat.setRotationDegrees(core::vector3df(relativeRotation.X, relativeRotation.Y, 0));
	mat.transformVect(target);

	mat.setRotationDegrees(core::vector3df(0, relativeRotation.Y, 0));
	mat.transformVect(movedir);

    movedir.normalize();
	
	if (eventReceiver.keyDown(KEY_KEY_W)) {
        if(oldNodePersonagemAnimationState != scene::EMAT_RUN)
            nodePersonagem->setMD2Animation(scene::EMAT_RUN);
        newNodePersonagemAnimationState = scene::EMAT_RUN;
        pos += movedir * frameDeltaTime * MoveSpeed;
        houveMovimento = true;
    }

	if (eventReceiver.keyDown(KEY_KEY_S)) {
		if(oldNodePersonagemAnimationState != scene::EMAT_RUN)
            nodePersonagem->setMD2Animation(scene::EMAT_RUN);
        newNodePersonagemAnimationState = scene::EMAT_RUN;
        pos -= movedir * frameDeltaTime * MoveSpeed;
        houveMovimento = true;
    }

	core::vector3df strafevect = target;
	strafevect = strafevect.crossProduct(camera->getUpVector());

	strafevect.Y = 0.0f;

	strafevect.normalize();

	if (eventReceiver.keyDown(KEY_KEY_A)) {
		if(oldNodePersonagemAnimationState != scene::EMAT_RUN)
            nodePersonagem->setMD2Animation(scene::EMAT_RUN);
        newNodePersonagemAnimationState = scene::EMAT_RUN;
        pos += strafevect * frameDeltaTime * MoveSpeed;
        houveMovimento = true;
    }

	if (eventReceiver.keyDown(KEY_KEY_D)) {
		if(oldNodePersonagemAnimationState != scene::EMAT_RUN)
            nodePersonagem->setMD2Animation(scene::EMAT_RUN);
        newNodePersonagemAnimationState = scene::EMAT_RUN;
        pos -= strafevect * frameDeltaTime * MoveSpeed;
		houveMovimento = true;
    }
    
    if(!houveMovimento) {
        nodePersonagem->setMD2Animation(scene::EMAT_STAND);
        newNodePersonagemAnimationState = scene::EMAT_STAND;
    }
    
	camera->setPosition(pos);
	nodePersonagem->setPosition(pos - core::vector3df(0,60,0));
	
    target += pos;
	camera->setTarget(target);
	
    nodePersonagem->setRotation(core::vector3df(0,relativeRotation.Y - 90,0));      

    return newNodePersonagemAnimationState;
}

/**
    Atualiza o movimento da partcula
*/

void atualizaBill(IBillboardSceneNode * bill, core::vector3df direcaoBill,u32 now, u32 then) {
    const f32 MoveSpeed = 200.0;
    const f32 frameDeltaTime = (f32)(now - then) / 1000.f;
    
    core::vector3df pos = bill->getPosition();
    
    pos += direcaoBill * frameDeltaTime * MoveSpeed;
    
    bill->setPosition(pos);
    
}

/**
    Calcula o vetor de reflexo de um vetor incidente e um vetor normal
*/

core::vector3df calculaReflexao(core::vector3df incidente, core::vector3df normal) {
    core::vector3df projecao;
    core::vector3df r;
    core::vector3df reflexao;
    
    projecao = incidente.dotProduct(normal) * normal;
    r = incidente - projecao;
    
    reflexao = -1 * incidente + 2 * r;
    
    return reflexao.normalize();
}

/**
    Decide se dois scene nodes colidem. til para calcular colises com os
portais.
    A bounding box  uma caixa 3D criada em volta do scene node, pela prpria
engine. Normalmente, ela  criada em torno da origem. O mtodo 
getTransformedBoundingBox a transporta da origem para ficar em volta do scene
node.
*/

bool collision( ISceneNode* sn1, ISceneNode* sn2 ) {
   return sn1->getTransformedBoundingBox().intersectsWithBox( sn2->getTransformedBoundingBox() );
} 

int main()
{
    // Cria o manipulador de eventos
    MasterEventReceiver eventReceiver;

    //Inicializa o dispositivo do Irrlicht
    
    IrrlichtDevice *device = createDevice(video::EDT_OPENGL, core::dimension2d<s32>(1024, 768), 16, false, false, false, &eventReceiver);
    
    if (device == 0)
       return 1; // no pde criar o driver do OPENGL
    
    video::IVideoDriver* driver = device->getVideoDriver();
    scene::ISceneManager* smgr = device->getSceneManager();
    
    //Carrega o mapa
    
    device->getFileSystem()->addZipFileArchive("map.pk3");
    
    scene::IAnimatedMesh* q3levelmesh = smgr->getMesh("nemesis.bsp");
    scene::ISceneNode* q3node = 0;
    
    if (q3levelmesh)
       q3node = smgr->addOctTreeSceneNode(q3levelmesh->getMesh(0));
    
    /* Cria os seletores de tringulos.
    O seletor do mundo guardar os tringulos que formam as paredes e o seletor da porta
guardar os tringulos que formam a porta. 
    O seletor meta agrupa vrios seletores.
    */
    
    scene::ITriangleSelector* selectorMundo = 0;
    scene::ITriangleSelector* selectorPorta;
    scene::IMetaTriangleSelector * meta = smgr->createMetaTriangleSelector();
    
    /* Coloca o mapa na origem e anexa o seletor do mundo nele. Este anexamento  s
    para que o seletor seja destruido quando o mapa for destruido */
    
    if (q3node)
    {
        selectorMundo = smgr->createOctTreeTriangleSelector(q3levelmesh->getMesh(0), q3node, 128);
        q3node->setTriangleSelector(selectorMundo);
    }
    
    // Cria a camera do jogador
    
    scene::ICameraSceneNode* cameraJogador = smgr->addCameraSceneNode(0,core::vector3df(0,44,0),core::vector3df(0, 0, 0),-1);

    // Cria a porta e o boto
    
    scene::IMeshSceneNode* porta = smgr->addCubeSceneNode(1.0,0,-1,core::vector3df(-255.8,175,65),core::vector3df(0,180,0),core::vector3df(0,162,130));
    scene::IMeshSceneNode* botao = smgr->addCubeSceneNode(1.0,0,-1,core::vector3df(-255.8,175,220),core::vector3df(0,180,0),core::vector3df(0,50,50));
    
    // Coloca os tringulos que formam a porta no seletor da porta
    
    selectorPorta = smgr->createTriangleSelectorFromBoundingBox(porta);

    // Adiciona os seletores no seletor meta
    
    meta->addTriangleSelector(selectorMundo);
    meta->addTriangleSelector(selectorPorta);
    
    selectorMundo->drop();

    /* Quando o model da Sydney  carregado na Irrlicht, ele estar virado de frente para o eixo x. Ento,
    setamos o lookat inicial da cmera nesta direo */
    
    core::vector3df lookAtInicial = cameraJogador->getPosition() + core::vector3df(100, 0, 0);
    cameraJogador->setTarget(lookAtInicial);
    
    /*Anexamos o animador de colises  cmera. O animador coloca gravidade na cmera e trata a coliso
    da mesma com as paredes. */
    
    if (meta)
    {
        scene::ISceneNodeAnimator* anim = smgr->createCollisionResponseAnimator(
        meta, cameraJogador, core::vector3df(30,60,30),
        core::vector3df(0,-0.5,0),
        core::vector3df(0,50,0));
        cameraJogador->addAnimator(anim);
        meta->drop();
        anim->drop();
    }
    
    // Cria a Sydney
    
    IAnimatedMesh* mesh = smgr->getMesh("sydney.md2");
	if (!mesh)
		return 1;
	IAnimatedMeshSceneNode* nodePersonagem = smgr->addAnimatedMeshSceneNode( mesh );
	
	// Aplica a textura no personagem e o torna invisvel
    
	if (nodePersonagem)
	{
		nodePersonagem->setMaterialFlag(EMF_LIGHTING, false);
		nodePersonagem->setMD2Animation(scene::EMAT_STAND);
		nodePersonagem->setMaterialTexture( 0, driver->getTexture("sydney.bmp") );
		nodePersonagem->setScale(core::vector3df(2.0,2.0,2.0));
        nodePersonagem->setVisible(false);
	}

    // Estados de animao para o personagem
    
    scene::EMD2_ANIMATION_TYPE oldNodePersonagemAnimationState = scene::EMAT_STAND;
    scene::EMD2_ANIMATION_TYPE newNodePersonagemAnimationState;
    
    // Cria a partcula e sua direo inicial
    
    scene::IBillboardSceneNode * bill = smgr->addBillboardSceneNode();
    bill->setMaterialType(video::EMT_TRANSPARENT_ADD_COLOR );
    bill->setMaterialTexture(0, driver->getTexture("particle.bmp"));
    bill->setMaterialFlag(video::EMF_LIGHTING, false);
    bill->setMaterialFlag(video::EMF_ZBUFFER, false);
    bill->setSize(core::dimension2d<f32>(20.0f, 20.0f));
    
    core::vector3df direcaoBill = core::vector3df(1.0,0.0,0.0);


	// Esconde o cursor do mouse
    device->getCursorControl()->setVisible(false);
    
    // Adiciona uma luz ao ambiente
	smgr->setAmbientLight(video::SColor(0,60,60,60));
    
    int lastFPS = -1;
    
    // Scene nodes dos portais
    
    PortalSceneNode *portalDireito = NULL;
    PortalSceneNode *portalEsquerdo = NULL;
    
    /*video::SMaterial material;
    material.setTexture(0, 0);
    material.Lighting = false;*/
    
    // Variveis utilizadas para os clculos do posicionamento, coliso e atravessamento do portal
    core::vector3df padrao_primeiro = core::vector3df(0.0,1.0,0.0);
    core::vector3df padrao_segundo = core::vector3df(0.0,0.0,-1.0);
    core::vector3df troca = core::vector3df(0.0,0.0,0.0);
    
    core::vector3df diretor_horizontal;
    core::vector3df diretor_vertical;
    core::vector3df normalPortalEsquerdo, normalPortalDireito, centroPortalEsquerdo, centroPortalDireito, trinormal;
    
    bool personagemAtravessouPortal = false;
    bool billAtravessouPortal = false;
    
    core::line3d<f32> lineColisao;
    core::vector3df intersection;
    core::triangle3df tri;
    
    core::line3d<f32> lineBill;
    core::vector3df intersectionBill;
    core::triangle3df triBill;
    
    // Varivel booleana indicando a troca entre os vetores diretores
    
    int fez_troca = 0;
    
    /* Cria a textura renderizvel de cada portal e a cmera para cada um deles.
    Os nomes "esquerdo" e "direito" servem para diferenciar o portal que  criado
    pelo boto esquerdo e o criado pelo boto direito do mouse.
    */
    video::ITexture* texturaPortalEsquerdo = 0;
    scene::ICameraSceneNode* cameraPortalEsquerdo = 0;
    texturaPortalEsquerdo = driver->addRenderTargetTexture(core::dimension2d<s32>(1024,1024), "RTT1");
    
    video::ITexture* texturaPortalDireito = 0;
    scene::ICameraSceneNode* cameraPortalDireito = 0;
    texturaPortalDireito = driver->addRenderTargetTexture(core::dimension2d<s32>(1024,1024), "RTT1");
    
    // Posiciona o cursor do mouse no centro da tela
    device->getCursorControl()->setPosition(0.5f, 0.5f);
    
    // Desenha a cena 3D carregada e ativa a cmera do jogador
    smgr->drawAll();
    smgr->setActiveCamera(cameraJogador);
    u32 then = device->getTimer()->getTime();
    
    while(device->run())
    {
        if (device->isWindowActive())
        {
            
            driver->beginScene(true, true, 0);
            
            /* Esta linha tem o mesmo objetivo da linha de colisao do personagem, mas
            detecta colises para a partcula */
            lineBill.start = bill->getPosition();
            lineBill.end = lineBill.start + direcaoBill * 1000.0f;
            
            // Detecta colises com os portais
            
            if(portalEsquerdo != NULL && portalDireito != NULL) 
            {
                // Personagem colidiu com o portal esquerdo
                if(collision(nodePersonagem, portalEsquerdo)) {
                    cameraJogador->setPosition(centroPortalDireito + normalPortalDireito * 2.0);
                    cameraJogador->setTarget(cameraJogador->getPosition() + normalPortalDireito * 100.0);
                    personagemAtravessouPortal = true;
                }
                // Partcula colidiu com o portal esquerdo
                if(collision(bill, portalEsquerdo)) {
                    bill->setPosition(centroPortalDireito + normalPortalDireito * 2.0);
                    direcaoBill = normalPortalDireito;
                }
            }
            
            // Coliso da partcula com o boto
            if(collision(bill, botao)) {
                porta->setPosition(core::vector3df(1000,1000,1000));
            }
            
            // Coliso da partcula com uma parede
            if (smgr->getSceneCollisionManager()->getCollisionPoint(lineBill, meta, intersectionBill, triBill)) {
                if(bill->getPosition().getDistanceFrom(intersectionBill) <= 5.0) {
                    direcaoBill = calculaReflexao(direcaoBill, triBill.getNormal().normalize());
                }
            }
            
            //smgr->drawAll();
            
            const u32 now = device->getTimer()->getTime(); 
            
            /* Se o personagem acabou de atravessar o portal, no atualizamos a posio da cmera.
            Isto  feito pois a funo de atualizao da cmera no sabe que ocorrou o teleporte.
             necessrio um meio melhor de corrigir isso */
            if(!personagemAtravessouPortal) {
                newNodePersonagemAnimationState = movimentaCamera(cameraJogador, smgr, eventReceiver, now, then, device->getCursorControl(), nodePersonagem, oldNodePersonagemAnimationState);
                oldNodePersonagemAnimationState = newNodePersonagemAnimationState;
            }
            personagemAtravessouPortal = false;
            
            atualizaBill(bill, direcaoBill,now, then);
            
            then = now;
            
            if(eventReceiver.GetMouseState().ButtonDown)
            {
                /* Desenhamos uma linha que comea na posio da cmera do jogador 
                e termina 1000 unidades  frente, na direo do target da cmera do jogador.
                Esta linha ser passada ao seletor de tringulos meta. */
                
                lineColisao.start = cameraJogador->getPosition();
                lineColisao.end = lineColisao.start + (cameraJogador->getTarget() - lineColisao.start).normalize() * 1000.0f;
                
                /* Testamos se o jogador clicou em uma parede. Os dois primeiros parmetros da
                funo getCollisionPoint so a linha de coliso e o seletor meta. O parmetro
                intersection ser atribudo pela funo e  o ponto da parede onde houve a
                coliso. Da mesma forma, o parmetro tri ser o tringulo da parede onde
                houve a coliso */
                
                if (smgr->getSceneCollisionManager()->getCollisionPoint(lineColisao, meta, intersection, tri))
                {
                    trinormal=tri.getNormal();
                    trinormal=trinormal.normalize();
                    
                    /* Para desenhar o portal, precisamos calcular seus vetores horizontal
                    e vertical, utilizando os vetores diretores. Caso o produto vetorial do
                    diretor vertical com a normal da parede resulte no vetor (0,0,0), isto
                    quer dizer que a parede  paralela ao eixo XZ. Ento, trocamos o vetor 
                    diretor vertical pelo horizontal e continuamos o clculo normalmente.
                    Marcamos a troca para podermos destrocar depois.*/
                    
                    diretor_vertical = padrao_primeiro.crossProduct(trinormal).normalize();
                    if(diretor_vertical.getLength() == 0.0)
                    {
                        fez_troca = 1;
                        troca = padrao_primeiro;
                        padrao_primeiro = padrao_segundo;
                        padrao_segundo = troca;
                        diretor_vertical = padrao_primeiro.crossProduct(trinormal).normalize();
                    }
                    
                    diretor_horizontal = diretor_vertical.crossProduct(trinormal).normalize();
                    
                    // Verficamos qual bnoto do mouse foi pressionado e criamos o portal correspondente
                    
                    if(eventReceiver.GetMouseState().LeftButtonDown) 
                    {
                        if(portalEsquerdo != NULL)
                            portalEsquerdo->remove();
                        
                        // Decidimos se o portal cabe na parede desejada
                        
                        if(portalFitsWall(intersection, tri, diretor_vertical, diretor_horizontal, 90, 120, driver, smgr, meta)) {
                            centroPortalEsquerdo = intersection;
                            normalPortalEsquerdo = trinormal;
                            portalEsquerdo = new PortalSceneNode(smgr->getRootSceneNode(), smgr, 666, tri, intersection, lineColisao, diretor_vertical, diretor_horizontal);
                            cameraPortalEsquerdo = smgr->addCameraSceneNode(0, intersection + tri.getNormal().normalize() * 5.0, intersection + tri.getNormal().normalize() * 10.0);
                        }
                    }
                    else 
                    {
                        if(portalDireito != NULL)
                            portalDireito->remove();
                        
                        // Decidimos se o portal cabe na parede desejada
                        
                        if(portalFitsWall(intersection, tri, diretor_vertical, diretor_horizontal, 90, 120, driver, smgr, meta)) {
                            centroPortalDireito = intersection;
                            normalPortalDireito = trinormal;
                            portalDireito = new PortalSceneNode(smgr->getRootSceneNode(), smgr, 666, tri, intersection, lineColisao, diretor_vertical, diretor_horizontal);
                            cameraPortalDireito = smgr->addCameraSceneNode(0, intersection + tri.getNormal().normalize() * 5.0, intersection + tri.getNormal().normalize() * 10.0);
                        }
                    }   
                    
                    if(fez_troca)
                    {
                        // destroca
                        troca = padrao_primeiro;
                        padrao_primeiro = padrao_segundo;
                        padrao_segundo = troca;
                        fez_troca = 0;
                    }
                }
                
                eventReceiver.resetMouseState();
                smgr->setActiveCamera(cameraJogador);
            }
            
            if(portalEsquerdo != NULL && portalDireito != NULL) 
            {
                /* Se os dois portais esto criados, renderizamos em um a viso da cmera
                do outro */
                
                portalEsquerdo->setMaterialTexture(0, texturaPortalEsquerdo);
                driver->setRenderTarget(texturaPortalEsquerdo, true, true, video::SColor(0,0,0,255));
                
                smgr->setActiveCamera(cameraPortalDireito);
                nodePersonagem->setVisible(true);
                portalEsquerdo->setVisible(false);
                smgr->drawAll();
                portalEsquerdo->setVisible(true);
                smgr->drawAll();
                smgr->drawAll();
                
                portalDireito->setMaterialTexture(0, texturaPortalDireito);
                driver->setRenderTarget(texturaPortalDireito, true, true, video::SColor(0,0,0,255));
                
                smgr->setActiveCamera(cameraPortalEsquerdo);
                portalDireito->setVisible(false);
                smgr->drawAll();
                portalDireito->setVisible(true);
                smgr->drawAll();
                smgr->drawAll();
                
                driver->setRenderTarget(0, true, true, 0);
                
                nodePersonagem->setVisible(false);
                
                smgr->setActiveCamera(cameraJogador);
            }
            
            smgr->drawAll();
            
            // Desenhamos a crosshair
            
            driver->draw2DRectangle(irr::video::SColor(100,255,100,100), irr::core::rect<s32>(driver->getScreenSize().Width / 2 - 15, driver->getScreenSize().Height / 2 - 1, driver->getScreenSize().Width / 2 + 15, driver->getScreenSize().Height / 2 + 1));
            driver->draw2DRectangle(irr::video::SColor(100,255,100,100), irr::core::rect<s32>(driver->getScreenSize().Width / 2 - 1, driver->getScreenSize().Height / 2 - 15, driver->getScreenSize().Width / 2 + 1, driver->getScreenSize().Height / 2 + 15));           

            driver->endScene();
            
            int fps = driver->getFPS();
            
            if (lastFPS != fps)
            {
                core::stringw str = L"PORTAL [";
                str += driver->getName();
                str += "] FPS:";
                str += fps;
                
                device->setWindowCaption(str.c_str());
                lastFPS = fps;
            }
        }
    }
    selectorMundo->drop();
    selectorPorta->drop();
    device->drop();
    
    return 0;
}
