#include <irrlicht.h>
#include <iostream>

using namespace irr;

// Scene node do portal

class PortalSceneNode : public scene::ISceneNode
{
    core::aabbox3d<f32> Box;
    video::S3DVertex Vertices[4];
    video::SMaterial Material;
    
    public:
    
    PortalSceneNode(scene::ISceneNode* parent, scene::ISceneManager* mgr, s32 id, core::triangle3df tri, core::vector3df intersection, core::line3d<f32> line, core::vector3df diretor_vertical, core::vector3df diretor_horizontal): scene::ISceneNode(parent, mgr, id)
    {
        /* Construtor do portal. Desenha os 4 vrtices do portal sabendo sua altura, largura
        e seus vetores diretores */
        
        Material.Wireframe = false;
        Material.Lighting = false;
        
        float width = 90.00;
        float height = 120.0;
        float dotcorrection = 0.2;
        core::vector3df trinormal;
        core::vector3df posicao_1_0;
        core::vector3df posicao_0_0;
        core::vector3df posicao_1_1;
        core::vector3df posicao_0_1;
        
        trinormal=tri.getNormal();
        
        trinormal=trinormal.normalize();
        
        posicao_1_0 = intersection + trinormal* dotcorrection - diretor_vertical * width/2 - diretor_horizontal * height/2;
        posicao_1_1 = intersection + trinormal* dotcorrection + diretor_horizontal * height/2 - diretor_vertical * width/2;
        posicao_0_1 = intersection + trinormal* dotcorrection + diretor_vertical * width/2 + diretor_horizontal * height/2;
        posicao_0_0 = intersection + trinormal* dotcorrection - diretor_horizontal * height/2 + diretor_vertical * width/2;
        
        Vertices[0] = video::S3DVertex(posicao_1_0.X, posicao_1_0.Y,posicao_1_0.Z,1,0,0,video::SColor(255,255,255,255),1,0);
        Vertices[1] = video::S3DVertex(posicao_1_1.X, posicao_1_1.Y,posicao_1_1.Z, 1,0,0,video::SColor(255,255,255,255),1,1);
        Vertices[2] = video::S3DVertex(posicao_0_1.X, posicao_0_1.Y,posicao_0_1.Z, 1,0,0,video::SColor(255,255,255,255),0,1);
        Vertices[3] = video::S3DVertex(posicao_0_0.X, posicao_0_0.Y,posicao_0_0.Z, 1,0,0,video::SColor(255,255,255,255),0,0);
        
        Box.reset(Vertices[0].Pos);
        for (s32 i=1; i<4; ++i)
            Box.addInternalPoint(Vertices[i].Pos);
    }
    
    virtual void OnRegisterSceneNode()
    {
            if (IsVisible)
                    SceneManager->registerNodeForRendering(this);

            ISceneNode::OnRegisterSceneNode();
    }

    virtual void render()
    {
            u16 indices[] = {       0,2,3, 2,1,3, 1,0,3, 2,0,1      };
            video::IVideoDriver* driver = SceneManager->getVideoDriver();

            driver->setMaterial(Material);
            driver->setTransform(video::ETS_WORLD, AbsoluteTransformation);
            driver->drawIndexedTriangleList(&Vertices[0], 4, &indices[0], 4);
    }

    virtual const core::aabbox3d<f32>& getBoundingBox() const
    {
            return Box;
    }

    virtual u32 getMaterialCount() const
    {
            return 1;
    }

    virtual video::SMaterial& getMaterial(u32 i)
    {
            return Material;
    }       
};
